class BaseError(Exception):
    """Base error class."""

    def __init__(self, message: str):
        self.message = message

    def __str__(self) -> str:
        return self.message


class InvalidOctalError(BaseError):
    """Error that represents an invalid octal entered by user."""

    def __init__(self, message: str):
        super().__init__(message=message)



import stat
from typing import Literal

from unix_perms._models import Authority

OWNER_PERMISSIONS = Authority(
    read_write_execute=stat.S_IRWXU,
    read_write=stat.S_IRUSR | stat.S_IWUSR,
    read=stat.S_IRUSR,
    read_execute=stat.S_IRUSR | stat.S_IXUSR,
    write_execute=stat.S_IWUSR | stat.S_IXUSR,
    write=stat.S_IWUSR,
    execute=stat.S_IXUSR,
)
GROUP_PERMISSIONS = Authority(
    read_write_execute=stat.S_IRWXG,
    read_write=stat.S_IRGRP | stat.S_IWGRP,
    read=stat.S_IRGRP,
    read_execute=stat.S_IRGRP | stat.S_IXGRP,
    write_execute=stat.S_IWGRP | stat.S_IXGRP,
    write=stat.S_IWGRP,
    execute=stat.S_IXGRP,
)
OTHERS_PERMISSIONS = Authority(
    read_write_execute=stat.S_IRWXO,
    read_write=stat.S_IROTH | stat.S_IWOTH,
    read=stat.S_IROTH,
    read_execute=stat.S_IROTH | stat.S_IXOTH,
    write_execute=stat.S_IWOTH | stat.S_IXOTH,
    write=stat.S_IWOTH,
    execute=stat.S_IXOTH,
)

PERMISSIONS_MAPPING = {
    "owner": OWNER_PERMISSIONS,
    "group": GROUP_PERMISSIONS,
    "others": OTHERS_PERMISSIONS,
}


class OctalPermissions:
    """
    An abstract interface facilitating access to octal permissions for
    read, write, and execute for any of the following authorites:
    ('owner', 'group', 'others').

    This class validates the given authority and assigns the corresponding
    permissions mapping.

    Args:
        authority (Literal['owner', 'group', 'others']): A specific permissions authority.

    Raises:
        ValueError: If 'authority' is not one of ('owner', 'group', 'others').
    """

    def __init__(self, authority: Literal["owner", "group", "others"]):
        self.authority = authority

        if authority not in {"owner", "group", "others"}:
            raise ValueError("Authority should be one of ('owner', 'group', 'others')")

        self._permissions: Authority = PERMISSIONS_MAPPING[authority]

    @property
    def no_permissions(self) -> int:
        return 0o000

    @property
    def read_write_execute(self) -> int:
        return self._permissions.read_write_execute

    @property
    def read_write(self) -> int:
        return self._permissions.read_write

    @property
    def read(self) -> int:
        return self._permissions.read

    @property
    def read_execute(self) -> int:
        return self._permissions.read_execute

    @property
    def write_execute(self) -> int:
        return self._permissions.write_execute

    @property
    def write(self) -> int:
        return self._permissions.write

    @property
    def execute(self) -> int:
        return self._permissions.execute



from pydantic import BaseModel


class Authority(BaseModel):
    """
    Represents octals for all Unix file permission settings for a specific authority.

    Args:
        read_write_execute (int): The octal for read, write, and execute permissions.
        read_write (int): The octal for read and write permissions.
        read (int): The octal for read only permissions.
        read_execute (int): The octal for read and execute permissions.
        write_execute (int): The octal for write and execute permissions.
        write (int): The octal for write only permissions.
        execute (int): The octal for execute only permissions.
    """

    read_write_execute: int
    read_write: int
    read: int
    read_execute: int
    write_execute: int
    write: int
    execute: int



from collections import namedtuple
from typing import Dict, Set, Union

from unix_perms._exceptions import InvalidOctalError

OctalConfig = namedtuple("OctalConfig", ["description", "read", "write", "execute"])

VALID_OCTAL_DIGITS: Set[str] = {str(num) for num in range(8)}

OCTAL_MODE_DIGIT_0 = OctalConfig(
    description="No permissions", read=False, write=False, execute=False
)
OCTAL_MODE_DIGIT_1 = OctalConfig(
    description="Execute permission only", read=False, write=False, execute=True
)
OCTAL_MODE_DIGIT_2 = OctalConfig(
    description="Write permission only", read=False, write=True, execute=False
)
OCTAL_MODE_DIGIT_3 = OctalConfig(
    description="Write and execute permissions", read=False, write=True, execute=True
)
OCTAL_MODE_DIGIT_4 = OctalConfig(
    description="Read permission only", read=True, write=False, execute=False
)
OCTAL_MODE_DIGIT_5 = OctalConfig(
    description="Read and execute permissions", read=True, write=False, execute=True
)
OCTAL_MODE_DIGIT_6 = OctalConfig(
    description="Read and write permissions", read=True, write=True, execute=False
)
OCTAL_MODE_DIGIT_7 = OctalConfig(
    description="Read, write, and execute permissions",
    read=True,
    write=True,
    execute=True,
)

OCTAL_DIGIT_CONFIGS: Dict[int, OctalConfig] = {
    0: OCTAL_MODE_DIGIT_0,
    1: OCTAL_MODE_DIGIT_1,
    2: OCTAL_MODE_DIGIT_2,
    3: OCTAL_MODE_DIGIT_3,
    4: OCTAL_MODE_DIGIT_4,
    5: OCTAL_MODE_DIGIT_5,
    6: OCTAL_MODE_DIGIT_6,
    7: OCTAL_MODE_DIGIT_7,
}


def _get_octal_digit_config(octal_digit: int) -> OctalConfig:
    """
    Private function to retrieve an OctalConfig object for a given octal digit.
    """
    if octal_digit in OCTAL_DIGIT_CONFIGS:
        return OCTAL_DIGIT_CONFIGS[octal_digit]
    else:
        raise InvalidOctalError(
            "Integer representation of an octal digit must be a single digit ranging from 0 to 7"
        )


def from_octal_digit_to_config(octal_digit: Union[str, int]) -> OctalConfig:
    """
    Given an octal digit, will return an OctalConfig object.

    Args:
        octal_digit (int): An integer octal digit, ranging from 0 to 7.

    Returns:
        OctalConfig: A named tuple containing basic permissions info for the digit.

    """
    if not isinstance(octal_digit, (str, int)):
        message_core = "Expected a string or integer object"
        raise TypeError(f"{message_core}, but got {type(octal_digit).__name__}")

    if isinstance(octal_digit, str):
        try:
            octal_digit = int(octal_digit)
        except ValueError:
            raise InvalidOctalError(
                "Expecting a string representation of an integer octal digit"
            )

    octal_config = _get_octal_digit_config(octal_digit=octal_digit)
    return octal_config


def _octal_validation(octal: str) -> str:
    """
    Private function to validate and convert a string Unix permissions mode
    to a three digit mode.
    """
    octal_string_length: int = len(octal)

    if not 1 <= octal_string_length <= 3:
        raise InvalidOctalError(
            "Invalid octal representation length, must have a length ranging from 0 to 3"
        )

    any_invalid_digits: bool = any(digit not in VALID_OCTAL_DIGITS for digit in octal)
    if any_invalid_digits:
        raise InvalidOctalError(
            "Invalid digits in octal representation, digits must range from 0 to 7"
        )

    octal_int_string_repr: str = octal.zfill(3)
    return octal_int_string_repr


def _from_decimal_to_permissions_mode(octal: int) -> str:
    """
    Private function to convert a decimal representation of an octal
    to a three digit string Unix permissions mode.
    """
    octal_string: str = format(octal, "o")
    octal_int_string_repr: str = _octal_validation(octal=octal_string)
    return octal_int_string_repr


def from_octal_to_permissions_mode(octal: Union[str, int]) -> str:
    """
    Creates a Unix permissions mode from an octal representation.

    This function accepts either a string or an integer as input. If the argument
    is a string, the value must be either in the format of an octal literal (e.g., '0o777')
    or as a Unix permissions mode (e.g., '777'). If the value is an integer, it
    must be a decimal representation of an octal as an octal literal (e.g., 0o777)
    or directly as an integer (e.g., 511).

    Args:
        octal (str | int): An octal representation as a string or integer.

    Returns:
        str: A string representation of a Unix permissions mode.
    """
    if not isinstance(octal, (str, int)):
        message_core = "Expected a string or integer object"
        raise TypeError(f"{message_core}, but got {type(octal).__name__}")

    permissions_mode: str
    if isinstance(octal, str):
        int_base = 10
        message = "Must be a valid decimal representation of an octal"

        if octal.startswith("0o"):
            int_base = 8
            message = "Must be a valid octal literal"

        try:
            octal_as_int = int(octal, int_base)
        except ValueError:
            raise InvalidOctalError(message)
        else:
            octal_as_str = str(octal_as_int)
            permissions_mode = _octal_validation(octal=octal_as_str)
            return permissions_mode
    else:
        permissions_mode = _from_decimal_to_permissions_mode(octal=octal)
        return permissions_mode


def is_permissions_mode(octal: Union[str, int]) -> bool:
    """
    A boolean function which determines if an octal representationis a valid
    Unix permissions mode.

    This function accepts either a string or an integer as input. If the argument
    is a string, the value must be either in the format of an octal literal (e.g., '0o777')
    or as a Unix permissions mode (e.g., '777'). If the value is an integer, it
    must be a decimal representation of an octal as an octal literal (e.g., 0o777)
    or directly as an integer (e.g., 511).

    Args:
        octal (str | int): An octal representation as a string or integer.

    Returns:
        bool: A boolean indicating whether the octal is a Unix permissions mode.
    """
    try:
        _ = from_octal_to_permissions_mode(octal=octal)
        is_perms_mode = True
    except InvalidOctalError:
        is_perms_mode = False
    return is_perms_mode



from __future__ import annotations

from typing import Dict, Literal, Optional, Union

from pydantic import BaseModel

from unix_perms._octals import (
    OctalConfig,
    from_octal_digit_to_config,
    from_octal_to_permissions_mode,
)
from unix_perms._permissions import OctalPermissions


class PermissionsConfig(BaseModel):
    """
    File permissions configuration for an octal digit. Can specify whether
    read, write, and execute permissions should be allowed.

    Args:
        read (bool): A boolean indicating whether read permission is included.
        write (bool): A boolean indicating whether write permission is included.
        execute (bool): A boolean indicating whether execute permission is included.
    """

    read: bool = True
    write: bool = False
    execute: bool = False

    @classmethod
    def from_octal_digit(cls, octal_digit: Union[str, int]) -> PermissionsConfig:
        """
        Creates a PermissionsConfig instance from an octal digit.

        Args:
            octal_digit (str | int): A string or integer octal digit, ranging
                from 0 to 7.

        Returns:
            PermissionsConfig: A new instance of PermissionsConfig corresponding
                to the octal digit.
        """
        octal_config: OctalConfig = from_octal_digit_to_config(octal_digit=octal_digit)
        return cls(
            read=octal_config.read,
            write=octal_config.write,
            execute=octal_config.execute,
        )


class PermissionsByte:
    """
    A simple structure representing a Unix permissions byte for one of any
    of the following authorites: ('owner', 'group', 'others').

    Args:
        authority (Literal['owner', 'group', 'others']): A specific permissions
            authority.
        config (PermissionsConfig | None): A file permissions configuration for
            an octal digit.
    """

    def __init__(
        self,
        authority: Literal["owner", "group", "others"],
        config: Optional[PermissionsConfig] = None,
    ):
        self.authority = authority
        self._config = PermissionsConfig() if config is None else config

        # Set permissions for specific authority
        self.permissions = OctalPermissions(authority=self.authority)

    def __add__(self, permission_byte: PermissionsByte) -> PermissionsMode:
        if not isinstance(permission_byte, PermissionsByte):
            raise TypeError(
                f'Can only add PermissionsByte (not "{type(permission_byte)}") '
                " to PermissionsByte"
            )

        return PermissionsMode._from_permissions_bytes(
            permissions_byte_one=self, permissions_byte_two=permission_byte
        )

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} authority={self.authority} "
            f"permissions_mode={self.permissions_mode}>"
        )

    @property
    def read_permission(self) -> bool:
        """A boolean indicating whether read permission is included."""
        return self._config.read

    @property
    def write_permission(self) -> bool:
        """A boolean indicating whether write permission is included."""
        return self._config.write

    @property
    def execute_permission(self) -> bool:
        """A boolean indicating whether execute permission is included."""
        return self._config.execute

    @property
    def permissions_mode(self) -> str:
        """The Unix permissions mode."""
        base_permissions = self.permissions.no_permissions
        if self._config.read:
            base_permissions |= self.permissions.read

        if self._config.write:
            base_permissions |= self.permissions.write

        if self._config.execute:
            base_permissions |= self.permissions.execute

        return format(base_permissions, "o").zfill(3)

    @property
    def permissions_description(self) -> str:
        """A string description of the Unix permissions mode."""
        permissions_mode = self.permissions_mode.strip("0")
        permissions_mode = permissions_mode or 0
        permissions_mode = int(permissions_mode)
        octal_config: OctalConfig = from_octal_digit_to_config(
            octal_digit=permissions_mode
        )
        return octal_config.description

    @property
    def permissions_description_detailed(self) -> Dict[str, Union[str, bool]]:
        """
        A more detailed description of the Unix permissions mode, as a dict.
        """
        return {
            "authority": self.authority,
            "mode": self.permissions_mode,
            "read": self.read_permission,
            "write": self.write_permission,
            "execute": self.execute_permission,
        }

    @property
    def permissions_mode_as_decimal_repr(self) -> int:
        """The decimal representation of the Unix permissions mode"""
        return int(self.permissions_mode, 8)

    @property
    def permissions_mode_as_int(self) -> int:
        """The integer representation of the Unix permissions mode"""
        return int(self.permissions_mode)

    @property
    def permissions_mode_as_octal_literal(self) -> str:
        """The string octal literal representation of the Unix permissions mode."""
        return f"0o{self.permissions_mode}"


_OCTAL_MAPPING: Dict[int, Literal["owner", "group", "others"]] = {
    0: "owner",
    1: "group",
    2: "others",
}
_CLASS_PARAMETERS = list(_OCTAL_MAPPING.values())


class PermissionsMode:
    """
    A simple structure representing a full Unix permissions mode, including
    all of the following authorites: ('owner', 'group', 'others').

    Args:
        owner (PermissionsByte): A PermissionsByte instance for the owner
            authority.
        group (PermissionsByte): A PermissionsByte instance for the group
            authority.
        others (PermissionsByte): A PermissionsByte instance for the others
        authority.
    """

    def __init__(
        self, owner: PermissionsByte, group: PermissionsByte, others: PermissionsByte
    ):
        self.owner = owner
        self.group = group
        self.others = others

    def __sub__(self, permission_byte: PermissionsByte) -> PermissionsMode:
        if not isinstance(permission_byte, PermissionsByte):
            raise TypeError(
                f'Can only subtract PermissionsByte (not "{type(permission_byte)}") '
                "from PermissionsMode"
            )

        permissions_mode: int = self.permissions_mode_as_decimal_repr
        permissions_mode_updated: int = (
            permissions_mode & ~permission_byte.permissions_mode_as_decimal_repr
        )
        class_arguments: Dict[str, PermissionsByte] = (
            PermissionsMode._transform_permission_mode(
                permissions_mode_updated=permissions_mode_updated
            )
        )

        return PermissionsMode(**class_arguments)

    def __add__(self, permission_byte: PermissionsByte) -> PermissionsMode:
        if not isinstance(permission_byte, PermissionsByte):
            raise TypeError(
                f'Can only add PermissionsByte (not "{type(permission_byte)}") '
                "to PermissionsMode"
            )

        permissions_mode: int = self.permissions_mode_as_decimal_repr
        permissions_mode_updated: int = (
            permissions_mode | permission_byte.permissions_mode_as_decimal_repr
        )
        class_arguments: Dict[str, PermissionsByte] = (
            PermissionsMode._transform_permission_mode(
                permissions_mode_updated=permissions_mode_updated
            )
        )

        return PermissionsMode(**class_arguments)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} permissions_mode={self.permissions_mode}>"

    def __str__(self) -> str:
        return repr(self)

    @staticmethod
    def _transform_permission_mode(
        permissions_mode_updated: int,
    ) -> Dict[str, PermissionsByte]:
        """
        Private static method to take the decimal representation of a
        transformed permissions mode from __add__ or __sub__, convert to
        the Unix permissions mode, and generate updated class arguments for
        a new PermissionsMode instance.
        """
        permission_mode: str = from_octal_to_permissions_mode(
            octal=permissions_mode_updated
        )
        class_arguments: Dict[str, PermissionsByte] = (
            PermissionsMode._generate_class_arguments_from_mode(
                permission_mode=permission_mode
            )
        )
        return class_arguments

    @staticmethod
    def _generate_class_arguments_from_mode(
        permission_mode: str,
    ) -> Dict[str, PermissionsByte]:
        """
        Private static method to generate class arguments for a PermissionsMode
        instance based on a Unix permission mode.
        """
        class_arguments: Dict[str, PermissionsByte] = dict()

        for index, octal_digit in enumerate(permission_mode):
            permissions_authority = _OCTAL_MAPPING[index]
            permissions_config = PermissionsConfig.from_octal_digit(
                octal_digit=octal_digit
            )

            permissions_byte = PermissionsByte(
                authority=permissions_authority, config=permissions_config
            )
            class_arguments.update({permissions_authority: permissions_byte})

        return class_arguments

    @classmethod
    def _from_permissions_bytes(
        cls,
        permissions_byte_one: PermissionsByte,
        permissions_byte_two: PermissionsByte,
    ) -> PermissionsMode:
        """
        Private class method to generate a PermissionsMode instance from
        two PermissionByte instances.
        """
        class_arguments: Dict[str, PermissionsByte] = dict()
        remaining_class_parameters = _CLASS_PARAMETERS.copy()

        if permissions_byte_one.authority == permissions_byte_two.authority:
            raise ValueError(
                "Authority cannot be the same for both PermissionsByte objects"
            )

        for byte in [permissions_byte_one, permissions_byte_two]:
            class_arguments.update({byte.authority: byte})
            remaining_class_parameters.remove(byte.authority)

        for authority in remaining_class_parameters:
            permissions_byte = PermissionsByte(authority=authority)
            class_arguments.update({authority: permissions_byte})

        return cls(**class_arguments)

    @classmethod
    def from_octal_representation(cls, octal: Union[str, int]) -> PermissionsMode:
        """
        Creates a PermissionsMode instance from an octal representation.

        This function accepts either a string or an integer as input. If
        the argument is a string, the value must be either in the format
        of an octal literal (e.g., '0o777') or as a Unix permissions mode
        (e.g., '777'). If the value is an integer, it must be a decimal
        representation of an octal as an octal literal (e.g., 0o777) or
        directly as an integer (e.g., 511).

        Args:
            octal (str | int): An octal representation as a string or integer.

        Returns:
            PermissionsMode: A new instance of PermissionsMode corresponding
                to the provided octal value.
        """
        permission_mode: str = from_octal_to_permissions_mode(octal=octal)
        class_arguments: Dict[str, PermissionsByte] = (
            PermissionsMode._generate_class_arguments_from_mode(
                permission_mode=permission_mode
            )
        )

        return cls(**class_arguments)

    @property
    def permissions_mode(self) -> str:
        """REturns the Unix permissions mode."""
        permissions_mode = (
            self.owner.permissions_mode_as_decimal_repr
            | self.group.permissions_mode_as_decimal_repr
            | self.others.permissions_mode_as_decimal_repr
        )
        return format(permissions_mode, "o").zfill(3)

    @property
    def permissions_mode_as_decimal_repr(self) -> int:
        """The decimal representation of the Unix permissions mode"""
        return int(self.permissions_mode, 8)

    @property
    def permissions_mode_as_int(self) -> int:
        """The integer representation of the Unix permissions mode"""
        return int(self.permissions_mode)

    @property
    def permissions_mode_as_octal_literal(self) -> str:
        """
        The string octal literal representation of the Unix permissions
        mode.
        """
        return f"0o{self.permissions_mode}"



from unix_perms._exceptions import InvalidOctalError
from unix_perms._octals import (
    OctalConfig,
    from_octal_digit_to_config,
    from_octal_to_permissions_mode,
    is_permissions_mode,
)
from unix_perms._permissions import OctalPermissions
from unix_perms._types import PermissionsByte, PermissionsConfig, PermissionsMode

__version__ = "0.6.0"
__all__ = [
    "InvalidOctalError",
    "OctalPermissions",
    "from_octal_digit_to_config",
    "from_octal_to_permissions_mode",
    "is_permissions_mode",
    "OctalConfig",
    "PermissionsByte",
    "PermissionsMode",
    "PermissionsConfig",
]
